;**********************************************************************
;                                                                     *
;    Filename:	    timer.asm                                         *
;    Date:                                                            *
;    File Version:                                                    *
;                                                                     *
;    Author:        el@jap.hu                                         *
;    Company:       http://jap.hu/                                    *
;**********************************************************************
;NOTES
;
;**********************************************************************
;HISTORY
;
; 000-20010602
; 001-20010603 working clock and timer, no display/keyboard/setting
; 002-20010603 keyboard handler OK
; 003-20010604 fully functional except output ctl
; 004-20010604 implementation done, write EEPROM routines
; 005-20010604 uses internal EEPROM to store on/off time
; 006-20010605 timing adjustment
; 007-20010706 bugfix: clear days at timer ON
;
;**********************************************************************
	list	p=16f84
	__CONFIG   _CP_OFF & _WDT_OFF & _PWRTE_ON & _HS_OSC

#include <p16F84.inc>

dcnt		EQU 0x0c ; digit counter
keys		equ 0x0d ; keyboard state
temp		equ 0x0e ; BCD temp
temp2
w_save		equ 0x10 ; itr save
s_save		equ 0x11 ; itr save
f_save		equ 0x12 ; itr save
porta_shadow	equ 0x13 ; porta mask for itr

flags		equ 0x14

digits		EQU 0x15 ; 7-segment display digits

; CLOCK
msec	equ	0x20		; tens of milliseconds
sec	equ	0x21		; seconds
min	equ	0x22		; minutes
hour	equ	0x23		; hours
day	equ	0x24		; days (max 32)

; START
smin	equ 0x25
shour	equ 0x26
sday	equ 0x27

; INTERVAL
imin	equ 0x28
ihour	equ 0x29

; don't change order of bytes between min-ihour!

; STOP
omin	equ 0x2a
ohour	equ 0x2b

r1	equ 0x2c		; register 1
r2	equ 0x2e		; register 2

mode	equ 0x30
prevkeys equ 0x31
keyidle	equ 0x32
keypress equ 0x33

dlycnt1	equ 0x34
dlycnt2	equ 0x35

; CONSTANTS

; FLAGS
ch_min	equ 0			; minutes changed
in_iv	equ 1			; check (ohour, omin) stop condition
al_ena	equ 2			; check (sday, shour, sday) start condition
ch_sec	equ 3			; seconds changed
ch_eep	equ 4			; permament data changed

; KEYS
KMODE	equ 7
KENA	equ 6
KSTART	equ 5
KSTOP	equ 4
KDAY	equ 6
KHOUR	equ 5
KMIN	equ 4

krdelay	equ .20
krrate	equ .20-3

; CLOCK

msinc_iv	equ .10000
msdiv		equ .100	; ms divider

; clkinc_iv = prescaler * 256 usec @ 4MHz
clkinc_iv	equ .4 * .256
msinc_step	equ .65536-msinc_iv

; DISPLAY
LMINUS		equ B'10000000'
LC		equ B'01110010'
LE		equ B'11110010'
LF		equ B'11100010'
LL		equ B'01110000'
LO		equ B'01111110'
LN		equ B'10101000'

; PORTS
PARELAY		equ 2

; EEPROM DATA
		ORG 0x2100
		de .22, .7, 0, 4, 0

vectors		ORG 0
  		goto main
		nop
		nop
		nop

		movwf w_save
		swapf STATUS, W
		movwf s_save
		movf FSR, W
		movwf f_save

		bcf INTCON, T0IF
		call interrupt

		movf f_save, W
		movwf FSR
		swapf s_save, W
		movwf STATUS
		swapf w_save
		swapf w_save, W
		retfie

segtable	bcf PCLATH, 0 ; on page 0
		addwf PCL, F
		dt B'01111110', B'00001100' ; 01
		dt B'10110110', B'10011110' ; 23
		dt B'11001100', B'11011010' ; 45
		dt B'11111010', B'00001110' ; 67
		dt B'11111110', B'11011110' ; 89
		dt B'11101110', B'11111000' ; Ab
		dt B'01110010', B'10111100' ; Cd
		dt B'11110010', B'11100010' ; EF

print_dec	movwf temp
		clrf	temp+1

pr_dec11	movlw	d'10'
		subwf	temp, W
		bnc	pr_dec1

		movwf	temp
		incf	temp+1, F
		goto	pr_dec11

pr_dec1		movf	temp+1, W
		call	print_chr
		movf temp, W

print_chr	call segtable
		movwf INDF
		incf FSR, F
		return

led_mpx		incf dcnt, F
		clrf PORTB
		movlw 0x09
		subwf dcnt, W
		bnz next_row

		; readout keys
		movf PORTB, W
		movwf keys

		; reset
		clrf dcnt
		movlw 3
		iorwf porta_shadow, W
		movwf PORTA
		movlw 1
		iorwf porta_shadow, W
		movwf PORTA
		movlw 0
		tris PORTB
		goto display

next_row	movlw 8
		subwf dcnt, W
		bnz notkeys
		movlw 0xff ; row 8:keyboard
		tris PORTB

notkeys		; clock
		movf porta_shadow, W
		movwf PORTA
		movlw 1
		iorwf porta_shadow, W
		movwf PORTA

display		movlw digits
		addwf dcnt, W
		movwf FSR
		movf INDF, W
		xorlw 0xff
		movwf PORTB

		movf porta_shadow, W
		btfsc INDF, 0
		movwf PORTA
		return

time_adj	; r1+= clkinc_iv
		movlw low clkinc_iv
		addwf r1, F
		btfsc STATUS, C
		incf r1+1, F
		movlw high clkinc_iv
		addwf r1+1, F

		; r2 = r1 + msinc_step
		movf r1+1, W
		movwf r2+1
		movlw low msinc_step
		addwf r1, W
		movwf r2
		btfsc STATUS, C
		incf r2+1, F
		movlw high msinc_step
		addwf r2+1, F

		; if (!overflow) return
		btfss STATUS, C
		return

		; r1 = r2
		movf r2, W
		movwf r1
		movf r2+1, W
		movwf r1+1

		incf	msec,F		;   msec++;
		movf	msec,W		;   ACCU = msec;
		sublw	msdiv		;   if ((ACCU-XD) != 0)
		btfss	STATUS,Z	;     return;
		return			;   else {
		clrf	msec		;     msec = 0;
		incf	sec,F		;     sec++;
		bsf flags, ch_sec	;     ch_sec = 1;
		movf	sec,W		;     ACCU = sec;
		sublw	0x3c		;     if ((ACCU-60) != 0)
		btfss	STATUS,Z	;	return;
		return			;     else {
		clrf	sec		;	sec = 0;
		incf	min,F		;	min++;
		bsf flags, ch_min	;     ch_min = 1;
		movf	min,W		;	ACCU = min;
		sublw	0x3c		;       if ((ACCU-60) != 0)
		btfss	STATUS,Z	;	  return;
		return			;       else {
		clrf	min		;	  min = 0;
		incf	hour,F		;	  hour++;
		movf	hour,W		;	  ACCU = hour;
		sublw	0x18		;         if ((ACCU-24) != 0)
		btfss	STATUS,Z	;	    return;
		return			;         else {
		clrf	hour		;          hour = 0;
		incf	day, F
		movf	day, W
		sublw	0x20
		btfsc	STATUS, Z
		clrf	day
		return

time_cmp	; compare time
		btfss flags, ch_min
		return	; no change in minutes

		bcf flags, ch_min
		btfss flags, in_iv
		goto skip_t1

		; interval end check
		movf ohour, W
		subwf hour, W
		btfss STATUS, Z
		goto skip_t1
		movf omin, W
		subwf min, W
		btfsc STATUS, Z
		bcf flags, in_iv
skip_t1
		; interval start check
		btfss flags, al_ena
		return

		; interval start check
		movf sday, W
		btfss STATUS, Z
		subwf day, W
		btfss STATUS, Z
		return
		movf shour, W
		subwf hour, W
		btfss STATUS, Z
		return
		movf smin, W
		subwf min, W
		btfss STATUS, Z
		return

time_start
		; start condition
		bsf flags, in_iv
		clrf day

		; recalc stop condition from (imin, ihour)
		movf ihour, W
		addwf hour, W
		movwf ohour

		movf imin, W
		addwf min, W
		movwf omin
		movlw .60
		subwf omin, W
		btfss STATUS, C
		goto skip_t2

		movwf omin
		incf ohour, F		

skip_t2		movlw .24
		subwf ohour, W
		btfsc STATUS, C
		movwf ohour
		return

main
		MOVLW 0
		MOVWF PORTA
		movwf porta_shadow
		movlw 0
		MOVWF PORTB

		movlw 0
		TRIS PORTA
		MOVLW 0
		TRIS PORTB

		; init display
		clrf dcnt

		; init clock
		clrf	r1
		clrf	r1+1
		clrf	msec		; msec = 0;
		clrf	sec		; sec = 0;
		clrf	min		; min = 0;
		clrf	hour		; hour = 0;
		clrf	day
		clrf	omin
		clrf	ohour
		clrf	flags
		call	eep_read

		; init mode
		clrf mode

		; init keys
		clrf keyidle
		clrf keypress
		movlw 0xff
		movwf prevkeys
		movwf keys

		; init interrupts
		clrwdt ; changing default presc. assignment
		movlw 0x01 ; prescaler 1:4 assigned to TMR0
		OPTION     ; T0CS selects internal CLK

		movlw (1<<GIE) + (1<<T0IE)
		movwf INTCON

loop		clrwdt
		call delay
		call time_cmp
		call key_handler
		call disp_update
		call relay_ctl
		call eep_write
		goto loop

interrupt	call time_adj
		goto led_mpx

delay		movlw 0x20
		movwf dlycnt2
delay2		nop
		decfsz dlycnt1, F
		goto delay2
		decfsz dlycnt2, F
		goto delay2
		return

key_handler	; handle keys

		movf keys, W
		xorwf prevkeys, W
		bnz keys_change

		incf keys, W
		bz keys_idle

keys_press	incf keypress, F
		goto keys_cont

keys_idle	incf keyidle, F
		goto keys_cont
		
keys_change	clrf keyidle
		clrf keypress
		movf keys, W
		movwf prevkeys

		; check for keys just pressed
		call keys_norepeat
		call keys_repeat

keys_cont	movlw krdelay
		subwf keypress, W
		bnc keys_c2

		; check for keys that repeat
		call keys_repeat
		movlw krrate
		movwf keypress

keys_c2		movlw 0xf0
		subwf keyidle, W
		bnc keys_c3

		; no key was pressed for long
		clrf mode
		clrf keyidle

keys_c3		; end of keyhandler
		return

keys_norepeat	btfsc keys, KMODE
		goto k_2

k_mode		; mode key was pressed
		incf mode, W
		andlw 0x03
		movwf mode
		return

k_2		movf mode, W
		btfss STATUS, Z
		return

		; norepeat keys in mode 0
		btfsc keys, KENA
		goto k_3

		; enable key in MODE0
		movlw 1<<al_ena
		xorwf flags, F
		clrf day
		return

k_3		; check for START & STOP
		btfsc flags, in_iv
		goto k_4

		; check for START
		btfsc keys, KSTART
		return
		goto time_start

k_4		; check for STOP
		btfss keys, KSTOP
		bcf flags, in_iv
		return

keys_repeat	movf mode, W
		btfsc STATUS, Z
		return

		; set hour&min pointer
		; FSR = 3xmode
		movwf FSR
		bcf STATUS, C
		rlf FSR, F
		addlw min-3
		addwf FSR, F

		btfsc keys, KHOUR
		goto k_6

		; hour key is pressed
		incf FSR, F
		incf INDF, F
		movlw .24
		subwf INDF, W
		btfsc STATUS, Z
		clrf INDF
		decf FSR, F
		call update_check

k_6		btfsc keys, KMIN
		goto k_7

		; minute key is pressed
		incf INDF, F
		movlw .60
		subwf INDF, W
		btfsc STATUS, Z
		clrf INDF
		call update_check

k_7		movlw 2
		subwf mode, W
		btfss STATUS, Z
		return

		; MODE 2, check day key
		btfsc keys, KDAY
		return

		incf sday, F
		clrf day
		movlw .31
		subwf sday, W
		btfsc STATUS, Z
		clrf sday

update_check	; set ch_eep flag if mode=1 or mode=2
		movlw 1
		xorwf mode, W
		btfsc STATUS, Z
		bsf flags, ch_eep

		xorlw 3
		btfsc STATUS, Z
		bsf flags, ch_eep

		return

disp_update	movlw digits
		movwf FSR

		bcf INTCON, GIE
		call disp_2

		; display relay state as a dot
		bcf digits+7, 0
		btfsc flags, in_iv
		bsf digits+7, 0

		bsf INTCON, GIE
		return

disp_2		movf mode, W
		bsf PCLATH, 0
		addwf PCL, F
		goto mode0
		goto mode1
		goto mode2

mode3		; display interval time
		movlw LO
		movwf digits
		movlw LF
		movwf digits+1
		movwf digits+2
		clrf digits+3

		movlw digits+4
		movwf FSR
		movf ihour, W
		call print_dec
		bsf digits+5, 0
		movf imin, W
		goto print_dec

mode0		; display time & on/off date
		movf hour, W
		call print_dec

		btfss sec, 0 ; display a flashing dot
		bsf digits+1, 0 ; dot
		movf min, W
		call print_dec

		btfsc flags, in_iv
		goto m0_started

		btfss flags, al_ena
		goto m0_noal

		; display start hour/min
		movf shour, W
		call print_dec
		bsf digits+5, 0 ; dot
		movf smin, W
		goto print_dec


m0_started	; display stop hour/min
		movf ohour, W
		call print_dec
		bsf digits+5, 0 ; dot
		movf omin, W
		goto print_dec

m0_noal		movlw LMINUS
		movwf digits+4
		movwf digits+5
		movwf digits+6
		movwf digits+7
		return
mode1		; display clock time
		movlw LC
		movwf digits
		movlw LL
		movwf digits+1
		clrf digits+2
		clrf digits+3
		movlw digits+4
		movwf FSR
		movf hour, W
		call print_dec
		bsf digits+5, 0
		movf min, W
		goto print_dec

mode2		; display start time
		movlw LO
		movwf digits
		movlw LN
		movwf digits+1
		incf FSR, F
		incf FSR, F
		movf sday, W
		call print_dec
		movf shour, W
		call print_dec
		bsf digits+5, 0
		movf smin, W
		goto print_dec

relay_ctl	; control output relay
		btfss flags, in_iv
		bcf porta_shadow, PARELAY
		btfsc flags, in_iv
		bsf porta_shadow, PARELAY
		return

eep_read	movlw smin
		movwf FSR
		clrf EEADR

eep_0		bcf INTCON, GIE
		bsf STATUS, RP0
		bsf EECON1, RD
		bcf STATUS, RP0
		bsf INTCON, GIE
		movf EEDATA, W
		movwf INDF

		incf FSR, F
		incf EEADR, F

		movlw ihour+1
		subwf FSR, W
		bnz eep_0
		return

eep_write	movf mode, W

		btfsc STATUS, Z
		btfss flags, ch_eep
		return

		bcf flags, ch_eep
		movlw LE
		movwf digits+4
		movwf digits+5
		clrf digits+6
		clrf digits+7

		; eeprom write cycle init
		movlw smin
		movwf FSR
		clrf EEADR

eep_1		movf INDF, W
		movwf EEDATA
		bcf INTCON, GIE
		bsf STATUS, RP0
		bcf EECON1, EEIF
		bsf EECON1, WREN
		movlw 0x55
		movwf EECON2
		movlw 0xaa
		movwf EECON2
		bsf EECON1, WR

		; wait for write completition
eep_2		bcf STATUS, RP0
		bsf INTCON, GIE
		nop
		nop
		bcf INTCON, GIE
		bsf STATUS, RP0
		btfss EECON1, EEIF
		goto eep_2
		bcf STATUS, RP0
		bsf INTCON, GIE

		incf FSR, F
		incf EEADR, F

		movlw ihour+1
		subwf FSR, W
		bnz eep_1
		return


		end

